-------------------------------------------------------------------------------
-- Concurrent Haskell Debugger 
--   DebugMsgChan Module
--     by Thomas Boettcher <thomas.boettcher@gmx.de>
-------------------------------------------------------------------------------

module CHD.DebugMsgChan
  (
    debugMsgChan,
    DebugMsg(..),
    DebugMsgAction(..),
    CHDInteractMsg(..), 
    reduceDebugMsg,
    interactCHD,
    msgCHD	 
  )
  where


-------------------------------------------------------------------------------
-- IMPORTS
-------------------------------------------------------------------------------

import qualified Control.Concurrent as C
import System.IO.Unsafe

import CHD.GuiWindow
import CHD.BaseTypes
import CHD.PriorDoubleChannel


-------------------------------------------------------------------------------
-- TYPES
-------------------------------------------------------------------------------

type DebugMsgChan = DblChan CHDInteractMsg DebugMessage


data CHDInteractMsg = 
    CHDQuit			-- Quitting Gui-Interface
  | CHDStop			-- Stop CHD (switching to Prio-Mode)
  | CHDContinue			-- Continue CHD (switching to DblChan)
  | Continue ThreadNo		-- Continue Thread
  | AutoContinue ThreadNo Int	-- AutoContinue with correct Identifier
  | Visualize Object Bool	-- set Visualization for Object
  | BreakAction DebugMsgAction BreakCounter -- set Break for Object
  | DisplayPosition String
  | GetInfoGuiWin GuiWindow
  deriving Show


-- complete message from the program to CHD

type DebugMessage = (C.ThreadId, DebugMsg, C.MVar ())


-- these two are needed to implement a protokoll-feature. 
-- this not happend yet!
-- one type cannot be written as string, the other does.

data DebugMsgAction = 
      ThreadBreakA		-- top Position because not setable

    | ProgramLabelA
    | ProgramStartA

    | ThreadForkSuspendA
    | ThreadForkA
    | ThreadKillSuspendA
    | ThreadKillA
    | ThreadStartA
    | ThreadEndA
    | ThreadLabelA
    | ThreadMarkA
    | ThreadUnmarkA
    | ThreadDelaySuspendA
    | ThreadDelayA
    
    | MVarNewEmptySuspendA
    | MVarNewEmptyA
    | MVarNewSuspendA
    | MVarNewA
    | MVarTakeSuspendA
    | MVarTakeA
    | MVarReadSuspendA
    | MVarReadA
    | MVarPutSuspendA
    | MVarPutA 
    | MVarSwapSuspendA
    | MVarSwapA
    | MVarTryTakeA
    | MVarTryPutA
    | MVarTryA
    | MVarDiedA
    | MVarLabelA

    | ChanNewSuspendA
    | ChanNewA
    | ChanWriteSuspendA
    | ChanWriteA
    | ChanReadSuspendA
    | ChanReadA
    | ChanUnGetSuspendA
    | ChanUnGetA
    | ChanDiedA
    | ChanLabelA

    | QSemNewSuspendA
    | QSemNewA
    | QSemWaitSuspendA
    | QSemWaitA
    | QSemSignalSuspendA
    | QSemSignalA
    | QSemDiedA
    | QSemLabelA

    | QSemNNewSuspendA
    | QSemNNewA
    | QSemNWaitSuspendA
    | QSemNWaitA
    | QSemNSignalSuspendA
    | QSemNSignalA
    | QSemNDiedA
    | QSemNLabelA

    | SampleVarNewEmptySuspendA
    | SampleVarNewEmptyA
    | SampleVarNewSuspendA
    | SampleVarNewA
    | SampleVarEmptySuspendA
    | SampleVarEmptyA
    | SampleVarReadSuspendA
    | SampleVarReadA
    | SampleVarWriteSuspendA
    | SampleVarWriteA
    | SampleVarDiedA 
    | SampleVarLabelA
  deriving (Show, Eq, Ord, Read, Enum)
instance Bounded DebugMsgAction where
  minBound = ProgramLabelA
  maxBound = SampleVarLabelA


-- main message-type.
--   implement new messages here first!

data DebugMsg = 
      ProgramLabel String
    | ProgramStart ReturnThread

    | ThreadStart
    | ThreadForkSuspend ReturnThread CodePosition
    | ThreadFork ReturnThread (ReturnMVar ()) CodePosition
    | ThreadKillSuspend ReturnThread CodePosition
    | ThreadKill ReturnThread CodePosition
    | ThreadEnd
    | ThreadLabel String (C.MVar String)
    | ThreadMark ReturnThread String
    | ThreadUnmark ReturnThread
    | ThreadDelaySuspend
    | ThreadDelay
    | ThreadBreak String
    
    | MVarNewEmptySuspend (ReturnMVar MVarNo) CodePosition
    | MVarNewEmpty MVarNo CodePosition
    | MVarNewSuspend (ReturnMVar MVarNo) CodePosition
    | MVarNew MVarNo String CodePosition
    | MVarTakeSuspend MVarNo CodePosition	-- will suspend on MVar
    | MVarTake MVarNo CodePosition		-- took from MVar
    | MVarReadSuspend MVarNo CodePosition	-- will suspend read on MVar
    | MVarRead MVarNo CodePosition		-- read from MVar
    | MVarPutSuspend MVarNo CodePosition	-- will put in MVar
    | MVarPut MVarNo String CodePosition	-- done put in MVar
    | MVarSwapSuspend MVarNo CodePosition	-- will suspend swap on MVar
    | MVarSwap MVarNo String CodePosition
    | MVarTryTake MVarNo CodePosition
    | MVarTryPut MVarNo CodePosition
    | MVarTry MVarNo CodePosition
    | MVarDied MVarNo				-- MVar is garbage collected
    | MVarLabel MVarNo String (C.MVar String)	-- label MVar

    | ChanNewSuspend (ReturnMVar ChanNo) CodePosition
    | ChanNew ChanNo CodePosition
    | ChanWriteSuspend ChanNo CodePosition
    | ChanWrite ChanNo String CodePosition
    | ChanReadSuspend ChanNo CodePosition
    | ChanRead ChanNo CodePosition
    | ChanUnGetSuspend ChanNo
    | ChanUnGet ChanNo
    | ChanDied ChanNo
    | ChanLabel ChanNo String (C.MVar String)

    | QSemNewSuspend (ReturnMVar QSemNo) Quantity CodePosition
    | QSemNew QSemNo CodePosition
    | QSemDied QSemNo
    | QSemWaitSuspend QSemNo CodePosition
    | QSemWait QSemNo CodePosition
    | QSemSignalSuspend QSemNo CodePosition
    | QSemSignal QSemNo CodePosition
    | QSemLabel QSemNo String (C.MVar String)

    | QSemNNewSuspend (ReturnMVar QSemNNo) Quantity CodePosition
    | QSemNNew QSemNNo CodePosition
    | QSemNDied QSemNNo
    | QSemNWaitSuspend QSemNNo CodePosition
    | QSemNWait Quantity QSemNNo CodePosition
    | QSemNSignalSuspend QSemNNo CodePosition
    | QSemNSignal Quantity QSemNNo CodePosition
    | QSemNLabel QSemNNo String (C.MVar String)

    | SampleVarNewEmptySuspend (ReturnMVar SampleVarNo) CodePosition
    | SampleVarNewEmpty SampleVarNo CodePosition
    | SampleVarNewSuspend (ReturnMVar SampleVarNo) CodePosition
    | SampleVarNew SampleVarNo String CodePosition
    | SampleVarEmptySuspend SampleVarNo CodePosition
    | SampleVarEmpty SampleVarNo CodePosition
    | SampleVarReadSuspend SampleVarNo CodePosition
    | SampleVarRead SampleVarNo CodePosition
    | SampleVarWriteSuspend SampleVarNo CodePosition
    | SampleVarWrite SampleVarNo String CodePosition
    | SampleVarDied SampleVarNo
    | SampleVarLabel SampleVarNo String (C.MVar String)

instance Show DebugMsg where
  showsPrec _ debugMsg =
    case debugMsg of 
      ProgramLabel name		    -> showString "label Program: " .
				       shows name
      ProgramStart _		    -> showString "start Program"

      ThreadForkSuspend _ _	    -> showString "will fork "
      ThreadFork _ _ _		    -> showString "forked "
      ThreadKillSuspend thread _    -> showString "killing " .
				       showsThread thread
      ThreadKill thread	_	    -> showString "killing " .
				       showsThread thread
      ThreadStart		    -> showString "START"
      ThreadEnd			    -> showString "END."
      ThreadLabel name _	    -> showString "label " . shows name
      ThreadMark thread _	    -> showString "mark " .
				       showsThread thread
      ThreadUnmark thread	    -> showString "unmark " .
				       showsThread thread
      ThreadDelaySuspend	    -> showString "delaying"
      ThreadDelay		    -> showString "running"
      ThreadBreak cause		    -> showString "chd-Break (". shows cause .
				       showString ")"

      MVarNewEmptySuspend no _	    -> showString "will NewEmptyMVar" .
				       showsNumber no
      MVarNewEmpty no _		    -> showString "NewEmptyMVar "
      MVarNewSuspend _ _	    -> showString "will NewMVar"
      MVarNew _ label _		    -> showString "NewMVar " . 
				       shows label
      MVarTakeSuspend no _	    -> showString "will TakeMVar " . shows no
      MVarTake no _		    -> showString "TakeMVar " . shows no
      MVarReadSuspend no _	    -> showString "will ReadMVar " . shows no
      MVarRead no _		    -> showString "ReadMVar " . shows no
      MVarPutSuspend no _	    -> showString "will PutMVar " . shows no
      MVarPut no label _	    -> showString "PutMVar " . shows no . 
				       shows label
      MVarSwapSuspend no _	    -> showString "will SwapMVar " . shows no
      MVarSwap no label _	    -> showString "SWAP " . shows no .
    				       showString label
      MVarTryTake no _		    -> showString "will TryTakeMVar " . 
				       shows no
      MVarTryPut no _		    -> showString "will TryPut " . shows no
      MVarTry no _ 		    -> showString "Take nothing from " .
				       shows no
      MVarDied no		    -> showString "removed MVar " . shows no .
				       showString " from Memory"
      MVarLabel no name	_	    -> shows no . showString " labeled " .
				       shows name


      ChanNewSuspend no	_	    -> showString "will NewChan" . 
				       showsNumber no
      ChanNew no _		    -> showString "creating new Channel" . 
				       shows no
      ChanWriteSuspend no _	    -> showString "will WRITE " . shows no
      ChanWrite no label _	    -> showString "WRITE " . shows no .
				       shows label
      ChanReadSuspend no _	    -> showString "suspend on READ " . shows no
      ChanRead no _		    -> showString "READ " . shows no
      ChanUnGetSuspend no	    -> showString "will UNGET " . shows no
      ChanUnGet no		    -> showString "UNGET " . shows no
      ChanDied no		    -> showString "removed Channel " .
				       shows no . showString " from Memory"
      ChanLabel no name _	    -> shows no . showString " labeled " .
				       shows name

			
      QSemNewSuspend no	_ _         -> showString "creating new QSem " .
				       showsNumber no	
      QSemNew no _		    -> showString "creating new QSem " .
				       shows no	
      QSemDied no		    -> showString "removed QSem " . shows no . 
				       showString " from Memory"
      QSemWaitSuspend no _	    -> showString "suspend on WAIT " . shows no
      QSemWait no _		    -> showString "WAIT " . shows no
      QSemSignalSuspend no _	    -> showString "suspend on SIGNAL " .
				       shows no
      QSemSignal no _		    -> showString "SIGNAL " . shows no
      QSemLabel no name	_	    -> shows no . showString " labeled " .
				       shows name


      QSemNNewSuspend no quantity _ -> showString "newQSemN" . showsNumber no .
				       shows "with Quantity" . shows quantity
      QSemNNew no _	            -> showString "creating new QSem " . 
				       shows no	
      QSemNDied no		    -> showString "removed QSemN " . shows no .
				       showString " from Memory"
      QSemNWaitSuspend no _	    -> showString "suspend on WAIT " . shows no
      QSemNWait _ no _		    -> showString "WAIT " . shows no
      QSemNSignalSuspend no _	    -> showString "suspend on SIGNAL " .
				       shows no
      QSemNSignal _ no _	    -> showString "SIGNAL " . shows no
      QSemNLabel no name _          -> shows no . showString " labeled " .
				       shows name


      SampleVarNewEmptySuspend no _ -> showString "will newEmptySampleVar" .
				       showsNumber no
      SampleVarNewEmpty no _	    -> showString "newEmptySampleVar" . 
				       shows no
      SampleVarNew no label _	    -> showString "creating new 
				       FullSampleVar " . shows no .
				       shows label   
      SampleVarEmptySuspend no _    -> showString "suspend on EMPTY " . 
				       shows no
      SampleVarEmpty no _	    -> showString "EMPTY " . shows no
      SampleVarReadSuspend no _     -> showString "suspend on READ " . shows no
      SampleVarRead no _	    -> showString "READ " . shows no
      SampleVarWriteSuspend no _    -> showString "suspend on WRITE " .
				       shows no
      SampleVarWrite no label _     -> showString "WRITE " . shows no .
				       showString label
      SampleVarDied no		    -> showString "removed SampleVar " .
				       shows no . showString " from Memory"
      SampleVarLabel no name _	    -> shows no . showString " labeled " .
				       shows name
      msg -> error (show msg)

    where 
    showsThread thread = 
      case thread of
        Number threadNo -> shows threadNo
        Identity _ -> showString "new Thread"

    showsNumber number =
      case number of
        CHD _ -> showString "(Number)"
        Gui num -> shows num


-------------------------------------------------------------------------------
-- FUNCTIONS
-------------------------------------------------------------------------------

reduceDebugMsg :: DebugMsg -> DebugMsgAction
reduceDebugMsg msg = 
  case msg of
    ProgramLabel _ -> ProgramLabelA
    ProgramStart _ -> ProgramStartA

    ThreadForkSuspend _ _ -> ThreadForkSuspendA
    ThreadFork _ _ _ -> ThreadForkA
    ThreadKillSuspend _ _ -> ThreadKillSuspendA
    ThreadKill _ _ -> ThreadKillA
    ThreadStart -> ThreadStartA
    ThreadEnd -> ThreadEndA
    ThreadLabel _ _ -> ThreadLabelA
    ThreadMark _ _ -> ThreadMarkA
    ThreadUnmark _ -> ThreadUnmarkA
    ThreadDelaySuspend -> ThreadDelaySuspendA
    ThreadDelay -> ThreadDelayA
    ThreadBreak _ -> ThreadBreakA

    MVarNewEmptySuspend _ _ -> MVarNewEmptySuspendA
    MVarNewEmpty _ _ -> MVarNewEmptyA
    MVarNewSuspend _ _ -> MVarNewSuspendA
    MVarNew _ _ _ -> MVarNewA
    MVarTakeSuspend _ _ -> MVarTakeSuspendA
    MVarTake _ _ -> MVarTakeA
    MVarReadSuspend _ _ -> MVarReadSuspendA
    MVarRead _ _ -> MVarReadA
    MVarPutSuspend _ _ -> MVarPutSuspendA
    MVarPut _ _ _ -> MVarPutA
    MVarSwapSuspend _ _ -> MVarSwapSuspendA
    MVarSwap _ _ _ -> MVarSwapA
    MVarTryPut _ _ -> MVarTryPutA
    MVarTryTake _ _ -> MVarTryTakeA
    MVarTry _ _ -> MVarTryA
    MVarDied _  -> MVarDiedA
    MVarLabel _ _ _ -> MVarLabelA

    ChanNewSuspend _ _ -> ChanNewSuspendA
    ChanNew _ _ -> ChanNewA
    ChanWriteSuspend _ _ -> ChanWriteSuspendA
    ChanWrite _ _ _ -> ChanWriteA
    ChanReadSuspend _ _ -> ChanReadSuspendA
    ChanRead _ _ -> ChanReadA
    ChanUnGetSuspend _ -> ChanUnGetSuspendA
    ChanUnGet _ -> ChanUnGetA
    ChanDied _ -> ChanDiedA
    ChanLabel _ _ _ -> ChanLabelA

    QSemNewSuspend _ _ _ -> QSemNewSuspendA 
    QSemNew _ _ -> QSemNewA 
    QSemWaitSuspend _ _ -> QSemWaitSuspendA
    QSemWait _ _ -> QSemWaitA
    QSemSignalSuspend _ _ -> QSemSignalSuspendA
    QSemSignal _ _ -> QSemSignalA
    QSemDied _ -> QSemDiedA
    QSemLabel _ _ _ -> QSemLabelA

    QSemNNewSuspend _ _ _ -> QSemNNewSuspendA
    QSemNNew _ _ -> QSemNNewA 
    QSemNWaitSuspend _ _ -> QSemNWaitSuspendA
    QSemNWait _ _ _ -> QSemNWaitA
    QSemNSignalSuspend _ _ -> QSemNSignalSuspendA
    QSemNSignal _ _ _ -> QSemNSignalA
    QSemNDied _ -> QSemNDiedA
    QSemNLabel _ _ _ -> QSemNLabelA

    SampleVarNewEmptySuspend _ _ -> SampleVarNewEmptySuspendA
    SampleVarNewEmpty _ _ -> SampleVarNewEmptyA
    SampleVarNewSuspend _ _ -> SampleVarNewSuspendA
    SampleVarNew _ _ _ -> SampleVarNewA
    SampleVarEmptySuspend _ _ -> SampleVarEmptySuspendA
    SampleVarEmpty _ _ -> SampleVarEmptyA
    SampleVarReadSuspend _ _ -> SampleVarReadSuspendA
    SampleVarRead _ _ -> SampleVarReadA
    SampleVarWriteSuspend _ _ -> SampleVarWriteSuspendA
    SampleVarWrite _ _ _ -> SampleVarWriteA    
    SampleVarDied _ -> SampleVarDiedA
    SampleVarLabel _ _ _ -> SampleVarLabelA


-------------------------------------------------------------------------------
-- DEBUGGER message channel constant
--   first time a Concurrent-function is called,
--   debugger will be started
--   (term will be reduced once)
-------------------------------------------------------------------------------

debugMsgChan :: DebugMsgChan
debugMsgChan = unsafePerformIO ( do
  dbgChan <- newDblChan
  myId <- C.myThreadId
  debugStop <- C.newEmptyMVar
  writeDblChanLess dbgChan (myId, ProgramStart (Identity myId), debugStop)
  return dbgChan
  )


-------------------------------------------------------------------------------
-- DEBUG message sender
--   create stopper to wait for ackknowledgement from user
-------------------------------------------------------------------------------

msgCHD :: DebugMsg -> IO (C.MVar ())
msgCHD message = do
  debugStop <- C.newEmptyMVar
  myId <- C.myThreadId
  writeDblChanLess debugMsgChan (myId, message, debugStop)
  return debugStop


interactCHD :: CHDInteractMsg -> IO ()
interactCHD message = do
  writeDblChanPrio debugMsgChan message
